﻿/* Jade Heliin
 * Paint By Number Creator
 * Order of functions:
 *  Open, Save, & Print
 *  Selecting buttons and coloring in regions
 *  Edge detection and creating labels
 *  Coloring, image processing algorithms
 */


using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Threading;

namespace PaintbyNumber
{
    public partial class MainWindow : Window
    {
        BitmapImage originalImage;      // original image selected by user
        string FILENAME;                // path of original image
        WriteableBitmap edgePBNImage;        // pbnImage with just edges

        bool colorButtonClicked = false;    // make sure color button is clicked before allowing click on image
        int colorButtonNumber;              // number of color button clicked
        int CELLSCOLOREDIN = 0;             // count of how many cells in image are colored in
        bool COLORINGDONE;                  // if image is all colored in or not
        bool[] regionColored;               // if a certain region is colored in or not

        // Struct of X and Y values for each pixel in image
        public struct Pixel
        {
            public int X;
            public int Y;
        }

        // Struct of X, Y, and cellnumbers for each location of label on image
        public struct LabelLocation
        {
            public int cellnumber;
            public int X;
            public int Y;
        }

        List<LabelLocation> Labels; // list of each label on image

        byte[,,] allEdgePixels; // y, x, and BGR for each pixel in edgePBNImage image

        List<List<int>> allNeighbourCells; // all neighbour cells of each cell

        // for flood fill
        int[,] allPixelCellNums;                // cellnumber for pixel at y,x
        byte[,,] allOriginalPixels;             // each pixel's y, x and BGR values
        List<List<Pixel>> allCells;             // list of allCells in image and all their allOriginalPixels
        List<List<Pixel>> finalAllCells;        // list of last round of cells before PMerging and all their allOriginalPixels
        byte[,] pixelMeanColor;                 // pixelMeanColor for pixel at y,x
        WriteableBitmap colorPBNImage;          // PBNImage with just colors
        int origPixelHeight;                    // height of originalImage
        int origPixelWidth;                     // width of originalImage
        
        int NumColorsWanted;  // number of colors wanted in image depending on complexity
        int NumCellsWanted;   // number of cells we want to be below, depending on complexity

        int FLOOD_FILL_TOLERANCE = 10;      // grouping allOriginalPixels into cells with color EucDistance < 10
        int SMALL_CELL_THRESHOLD = 30;      // merge cells with < 30 allOriginalPixels
        int CLOSE_CELL_TOLERANCE = 5;       // merge cells with pixelMeanColor EucDistance < 5
        double FIRST_PASS_N_RATIO = 1.5;    // first pass throough NMerging cell
        int K_MEANS_TRIALS = 30;            // how many times to go through PMerging for best P colors

        int VISITEDPIXEL = 1;           // if pixel is visited
        int UNVISITEDPIXEL = 0;         // if pixel is unvisited
        int PIXELNOTINCELL = -1;        // if pixel not in cell yet
        int CurrCellCount;              // current amount of cells
        //int CurrColorCount;             // current amount of colors

        ProgressWindow window;  // new window for progress bar during image processing

        int saveImageNumber = 0;    // number to save each image in desktop folder


        // to save intermediate images for presentations
        SaveFileDialog save_file = new SaveFileDialog
        {
            Title = "Save As",
            Filter = "All supported graphics|*.jpg;*.jpeg;*.png|" +
                        "jpeg (*.jpg;*.jpeg)|*.jpg;*.jpeg|" +
                        "portable network graphic (*.png)|*.png"
        };

        PngBitmapEncoder enc;

        // file path for each saved image
        string saveFilePath = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        
        string complexity;  // user's selected complexity option

        int[] cellMapCellNumber;  // maps cell to it's new cellnumber



        /*******************************************************************************
         * Reset everything, disabling and enabling, opening user's selected image   */
        private void OpenImageClick(object sender, RoutedEventArgs e)
        {
            PBNCanvas.IsEnabled = true;       // enable clicking on canvas
            save.IsEnabled = false;         // enable saving image
            print.IsEnabled = false;        // enable printing image

            
            // Need to unregister names of labels on image so I can reregister them
            int textNumber = 0;
            for (int i = 1; i < PBNCanvas.Children.Count; i++)
            {
                PBNCanvas.UnregisterName("text" + textNumber);
                textNumber++;
            }

            // Remove all labels from the image
            if (PBNCanvas.Children.Count > 1)
                PBNCanvas.Children.RemoveRange(1, PBNCanvas.Children.Count - 1);
            
            // Disable all buttons so can't be clicked
            foreach (Button btn in buttonGrid.Children)
                btn.IsEnabled = false;


            // Open file dialog for user to select picture to use
            OpenFileDialog openFileDialog = new OpenFileDialog
            {
                Title = "Select an Image:",
                Filter = "All supported graphics|*.jpg;*.jpeg;*.png|" +
                "jpeg (*.jpg;*.jpeg)|*.jpg;*.jpeg|" +
                "portable network graphic (*.png)|*.png " //|" + 
                //"tiff (*.tiff)|*.tiff|" + 
                //"originalImage (*.bmp)|*.bmp"
            };


            // If image is chosen, set originalImage and enable createPBN menu item
            if (openFileDialog.ShowDialog() == true)
            {
                originalImage = new BitmapImage();
                originalImage.BeginInit();
                originalImage.UriSource = new Uri(openFileDialog.FileName);
                originalImage.EndInit();

                // Get filename for later use
                FILENAME = Path.GetFileNameWithoutExtension(openFileDialog.FileName);

                // Set image on screen to original image and enable item
                PBNImage.Source = originalImage;
                createPBN.IsEnabled = true;
            }
        }


        /******************************************************************************
         * Save current PBNImage on canvas to user's selected path                   */
        private void SaveImageClick(object sender, RoutedEventArgs e)
        {
            // Save dialog to allow user to save current canvas
            SaveFileDialog save_file = new SaveFileDialog
            {
                Title = "Save As",
                Filter = "All supported graphics|*.jpg;*.jpeg;*.png|" +
                "jpeg (*.jpg;*.jpeg)|*.jpg;*.jpeg|" +
                "portable network graphic (*.png)|*.png"
            };

            // If user selected good path and name
            if (save_file.ShowDialog() == true)
            {
                // Create new bitmap with size of PBNImage on canvas
                //  and save to path
                Size size = new Size(this.Width, this.Height);
                PBNCanvas.Measure(size);

                var rtb = new RenderTargetBitmap(
                    (int)this.Width,                // width 
                    (int)this.Height,               // height 
                    96,                             // dpiX 
                    96,                             // dpiY 
                    PixelFormats.Default            // pixelformat 
                    );

                rtb.Render(PBNCanvas);

                var enc = new PngBitmapEncoder();
                enc.Frames.Add(BitmapFrame.Create(rtb));
                using (var stm = System.IO.File.Create(save_file.FileName))
                    enc.Save(stm);
            }
        }


        /*****************************************************************************
         * Print current PBNImage on canvas to user's selected printer              */
        private void PrintImageClick(object sender, RoutedEventArgs e)
        {
            // Create new print dialog
            PrintDialog printDlg = new PrintDialog();
            printDlg.PageRangeSelection = PageRangeSelection.AllPages;
            printDlg.UserPageRangeEnabled = true;

            if (printDlg.ShowDialog() == true)
            {
                // Save as in SaveImageClick, create new bitmap from canvas and print
                Size size = new Size(this.Width, this.Height);
                PBNCanvas.Measure(size);

                var rtb = new RenderTargetBitmap(
                    (int)this.Width,                // width 
                    (int)this.Height,               // height 
                    96,                             // dpiX
                    96,                             // dpiY 
                    PixelFormats.Default            // pixelformat 
                );

                rtb.Render(PBNCanvas);

                var enc = new PngBitmapEncoder();
                enc.Frames.Add(BitmapFrame.Create(rtb));

                printDlg.PrintVisual(PBNCanvas, "Print Paint by Number");
            }
        }


        /***************************************************************************
         * Handler for color button clicks, get button and and change color       */
        private void ColorButton_Click(object sender, RoutedEventArgs e)
        {
            // reset previous button if clicked
            if (colorButtonClicked)
            {
                Button prevButton = PBNCanvas.FindName("btn" + colorButtonNumber) as Button;

                // change back to color of region in image
                prevButton.Background = new SolidColorBrush(Color.FromArgb(255, pixelMeanColor[colorButtonNumber, 2], pixelMeanColor[colorButtonNumber, 1], pixelMeanColor[colorButtonNumber, 0]));
                prevButton.Background.Opacity = 0.80;
            }

            colorButtonClicked = true;
            Button b = sender as Button;
            colorButtonNumber = int.Parse(b.Content.ToString());

            // set current button chosen to white
            SolidColorBrush brush = new SolidColorBrush(Colors.White);
            b.Background.Opacity = 0.80;
            b.BorderBrush = Brushes.Black;
            b.Background = brush;
        }


        /****************************************************************************************
         * Handler for clicking on cells in edgePBNImage, color in cell if correct button clicked   */
        private void ColorRegion_Click(object sender, RoutedEventArgs e)
        {
            // steps:
            // only do this code if one color has first been clicked in the beginning
            // only color it in if the cellnumber of button matches the cellnumber of region
            if (COLORINGDONE == false)
            {
                if (colorButtonClicked)
                {
                    double x = Mouse.GetPosition(PBNImage).X;
                    double y = Mouse.GetPosition(PBNImage).Y;

                    Size renderSize = new Size(PBNImage.ActualWidth, PBNImage.ActualHeight);
                    int newx = (int)(x / (renderSize.Width / PBNImage.Source.Width));
                    int newy = (int)(y / (renderSize.Height / PBNImage.Source.Height));

                    if (newx >= 0 && newy >= 0 && newx < origPixelWidth && newy < origPixelHeight)
                    {
                        int finalCellNumber = allPixelCellNums[newy, newx];

                        if (finalCellNumber == colorButtonNumber)
                        {
                            byte[] pix = new byte[3];
                            for (int i = 0; i < 3; i++)
                                pix[i] = pixelMeanColor[finalCellNumber, i];

                            // Find cell with this pixel
                            Pixel n = new Pixel();
                            n.X = newx;
                            n.Y = newy;
                            //int listCellNumber = allCells.FindIndex(xx => xx.Contains(n));
                            int listCellNumber = finalAllCells.FindIndex(xx => xx.Contains(n));

                            if (regionColored[listCellNumber] == false)
                            {
                                regionColored[listCellNumber] = true;
                                CELLSCOLOREDIN++;

                                // remove cellnumber from region , add one from image
                                TextBlock txt = PBNCanvas.FindName("text" + listCellNumber) as TextBlock;
                                PBNCanvas.Children.Remove(txt);

                                foreach (Pixel p in finalAllCells[listCellNumber])
                                {
                                    int yy = p.Y;
                                    int xx = p.X;

                                    allEdgePixels[yy, xx, 0] = pixelMeanColor[allPixelCellNums[yy, xx], 0];
                                    allEdgePixels[yy, xx, 1] = pixelMeanColor[allPixelCellNums[yy, xx], 1];
                                    allEdgePixels[yy, xx, 2] = pixelMeanColor[allPixelCellNums[yy, xx], 2];

                                    allEdgePixels[yy, xx, 3] = 255;

                                }

                                byte[] pixels1D = new byte[origPixelHeight * origPixelWidth * 4];
                                int ind = 0;
                                for (int yy = 0; yy < origPixelHeight; yy++)
                                {
                                    for (int xx = 0; xx < origPixelWidth; xx++)
                                    {
                                        for (int i = 0; i < 4; i++)
                                            pixels1D[ind++] = allEdgePixels[yy, xx, i];
                                    }
                                }

                                Int32Rect rect = new Int32Rect(0, 0, origPixelWidth, origPixelHeight);
                                int str = 4 * origPixelWidth;
                                colorPBNImage.WritePixels(rect, pixels1D, str, 0);

                                PBNImage.Source = colorPBNImage;

                                if (CELLSCOLOREDIN == finalAllCells.Count)
                                    COLORINGDONE = true;
                            }
                        }
                    }
                }
            }
        }

        
        /********************************************************************************
         * Create edgePBNImage using edge detection on colorPBNImage and putting labels on image */
        private void EdgeDetect_Click()
        {
            COLORINGDONE = false;
            regionColored = new bool[finalAllCells.Count];
            for (int i = 0; i < regionColored.Count(); i++)
                regionColored[i] = false;
            
            edgePBNImage = new WriteableBitmap(origPixelWidth, origPixelHeight, 96, 96, PixelFormats.Bgra32, null);
            
            allEdgePixels = new byte[origPixelHeight, origPixelWidth, 4];
            
          
            for (int y = 0; y < origPixelHeight; y++)
            {
                for (int x = 0; x < origPixelWidth; x++)
                {
                    int value = FindEdge(allOriginalPixels, y, x);
                    for (int i = 0; i < 4; i++)
                        allEdgePixels[y, x, i] = (byte)value;
                    allEdgePixels[y, x, 3] = 255;
                }
            }
            

            byte[] pixels1D = new byte[origPixelHeight * origPixelWidth * 4];
            int ind = 0;
            for (int y = 0; y < origPixelHeight; y++)
            {
                for (int x = 0; x < origPixelWidth; x++)
                {
                    for (int i = 0; i < 4; i++)
                        pixels1D[ind++] = allEdgePixels[y, x, i];
                }
            }

            Int32Rect rect = new Int32Rect(0, 0, origPixelWidth, origPixelHeight);
            int str = 4 * origPixelWidth;
            edgePBNImage.WritePixels(rect, pixels1D, str, 0);
        
            PBNImage.Source = edgePBNImage;
            


            // Get label locations, and create textblocks to put over image on canvas
            LabelLocations();
            TextBlock txt;
            int newX, newY;
            Size renderSize = new Size(PBNImage.ActualWidth, PBNImage.ActualHeight);

            int textNumber = 0;
            foreach (LabelLocation labelloc in Labels)
            {

                txt = new TextBlock() {
                    Text = labelloc.cellnumber.ToString(),
                    Foreground = Brushes.Gray,
                    Background = Brushes.Transparent,
                    FontWeight = FontWeights.Bold,
                    //HorizontalAlignment = HorizontalAlignment.Left,
                    //VerticalAlignment = VerticalAlignment.Center,
                    
                };
                
                newX = (int)(labelloc.X / (PBNImage.Source.Width / renderSize.Width));
                newY = (int)(labelloc.Y / (PBNImage.Source.Height / renderSize.Height));
                
                newX = (int)(newX + (PBNImage.DesiredSize.Width - PBNImage.ActualWidth) / 2);
                newY = (int)(newY + (PBNImage.DesiredSize.Height - PBNImage.ActualHeight) / 2);

                Canvas.SetLeft(txt, newX);
                Canvas.SetTop(txt, newY - 6);

                MouseUp += new MouseButtonEventHandler(ColorRegion_Click);

                // check if name registered
                var text = PBNCanvas.FindName("text" + textNumber);
                if (text != null)
                    UnregisterName("text" + textNumber);
                RegisterName("text" + textNumber, txt);
                
                PBNCanvas.Children.Add(txt);
                textNumber++;
            }
            
        }
        public int FindEdge(byte[,,] pixels, int y, int x)
        {
            int myCellNum = allPixelCellNums[y, x];
            int[] xRel = new int[] { 1, 0 };
            int[] yRel = new int[] { 0, 1 };
            
            int xx, yy;
            for (int i = 0; i < xRel.Length; i++)
            {
                xx = x + xRel[i];
                yy = y + yRel[i];

                if (xx >= 0 && xx < origPixelWidth && yy >= 0 && yy < origPixelHeight)
                {
                    if (xx == 0 || xx == origPixelWidth - 1 || yy == 0 || yy == origPixelHeight - 1)
                        return 135;

                    Pixel p = new Pixel();
                    p.X = xx;
                    p.Y = yy;

                    if (allPixelCellNums[yy, xx] != myCellNum)
                        return 135;
                }
            }
            return 255;
        }


        /****************************************************************
         * Get label locations for each cell in image                  */
        public void LabelLocations()
        {
            Labels = new List<LabelLocation>();

            foreach (List<Pixel> cell in finalAllCells) //allCells)
            {
                Labels.Add(GetLabelLocation(cell));
            }
        }
        public LabelLocation GetLabelLocation(List<Pixel> cell)
        {
            int bestI = 0;
            int best = 0;
            int goodness;
            for (int i = 0; i < cell.Count; i++)
            {
                int x = cell[i].X;
                int y = cell[i].Y;
                goodness = SameCount(x, y, -1, 0) * SameCount(x, y, 1, 0) * SameCount(x, y, 0, -1) * SameCount(x, y, 0, 1);

                if (goodness > best)
                {
                    best = goodness;
                    bestI = i;
                }
            }

            LabelLocation labelloc = new LabelLocation();
            labelloc.cellnumber = allPixelCellNums[cell[0].Y, cell[0].X];
            labelloc.X = cell[bestI].X;
            labelloc.Y = cell[bestI].Y;

            return labelloc;

        }
        public int SameCount(int x, int y, int incX, int incY)
        {
            int cellnumber = allPixelCellNums[y, x];
            int count = -1;

            while (x >= 0 && x < origPixelWidth && y >= 0 && y < origPixelHeight && cellnumber == allPixelCellNums[y, x])
            {
                count++;
                x += incX;
                y += incY;
            }
            return count;
        }

        
        /*************************************************************
         * Creates window for user to selected complexity           */
        private void CreateComplexityDialog()
        {
            ComplexityDialog complexDlg = new ComplexityDialog();
            if (complexDlg.ShowDialog() == true)
                complexity = complexDlg.Complexity; // returns user's selected complexity, beginner if none selected
        }


        /********************************************************************
         * Create thread and new window for progress bar                   */
        private void ProgressBarStart()
        {
            Thread thread = new Thread(new ThreadStart(ThreadStartingPoint));
            thread.SetApartmentState(ApartmentState.STA);
            thread.IsBackground = true;
            thread.Start(); 
        }
        private void ThreadStartingPoint()
        {
            window = new ProgressWindow();
            window.Show();  // starts progress bar automatically
            Dispatcher.Run();
        }


        /****************************************************************************************************
         * Create colorPBNImage from originalImage using FloodFill and merging algorithms                       */
        private void CreatePBN_Click(object sender, RoutedEventArgs e)
        {

            
            // Save Intermediate Images
            saveImageNumber = 0;
            enc = new PngBitmapEncoder();
            enc.Frames.Add(BitmapFrame.Create(PBNImage.Source as BitmapSource));
            using (var stm = System.IO.File.Create(saveFilePath + FILENAME + saveImageNumber++ + ".png"))
                enc.Save(stm);
            


            // enable & disable buttons
            createPBN.IsEnabled = false;
            save.IsEnabled = true;
            print.IsEnabled = true;


            /********************************************************************************/
            // Have user choose difficulty, and set NumCellsWanted number of cells and NumColorsWanted number of colors
            CreateComplexityDialog();
            
            if (complexity == "Beginner")
            {
                NumColorsWanted = 5;
                NumCellsWanted = 40;
                
            }
            else if (complexity == "Intermediate")
            {
                NumColorsWanted = 10;
                NumCellsWanted = 70;
            }
            else // complexity == "Advanced"
            {
                NumColorsWanted = 14;
                NumCellsWanted = 100;
            }
            

            
            /************************************************/
            // Create background thread to show progress bar
            ProgressBarStart();
           
        
            // Start processing image, create PBN
            StartCreatingPBN();



            /*
            // Save Intermediate Images
            enc = new PngBitmapEncoder();
            enc.Frames.Add(BitmapFrame.Create(PBNImage.Source as BitmapSource));
            using (var stm = System.IO.File.Create(saveFilePath + FILENAME + saveImageNumber++ + ".png"))
                enc.Save(stm);
            */
            
            
            // Put Colors in Buttons
            Button button;
            SolidColorBrush brush;

            int B, G, R;
            string[] colorLines = new string[NumColorsWanted];  // save colors to write to file
            for (int i = 0; i < NumColorsWanted; i++)
            {
                // get button with same cellnumber and set it's color and properties
                button = PBNCanvas.FindName("btn" + i) as Button;
                B = pixelMeanColor[i, 0];
                G = pixelMeanColor[i, 1];
                R = pixelMeanColor[i, 2];
                brush = new SolidColorBrush(Color.FromArgb(255, pixelMeanColor[i, 2], pixelMeanColor[i, 1], pixelMeanColor[i, 0]));
                
                button.Background = brush;
                button.Background.Opacity = 0.80;
                button.IsEnabled = true;
                button.Visibility = Visibility.Visible;
                
                // save color
                colorLines[i] = "Cell " + i + ": B - " + B + " G - " + G + " R - " + R;
            }

            // for color buttons that aren't being used, set to hidden
            for (int i = NumColorsWanted; i < buttonGrid.Children.Count; i++)
            {
                button = buttonGrid.Children[i] as Button;
                button.Visibility = Visibility.Hidden;
            }

            // write colors to file
            using (StreamWriter outputFile = new StreamWriter(Path.Combine(saveFilePath, FILENAME + "Colors.txt")))
            {
                foreach (string color in colorLines)
                    outputFile.WriteLine(color);
            }

            
            // start edge detection on colored image, creating edgePBN
            EdgeDetect_Click();


            // stop running progress bar
            window.Dispatcher.Invoke(DispatcherPriority.Normal, new ThreadStart(window.Close));
        }


        /*****************************************************************************
         * Start creating PBN process                                               */
         private void StartCreatingPBN()
        {
            origPixelWidth = originalImage.PixelWidth;
            origPixelHeight = originalImage.PixelHeight;
            
            // 1D array of all pixels, will need to copy back into to put back into image
            byte[] originalPixels = new byte[origPixelHeight * origPixelWidth * 4];
            int stride = 4 * origPixelWidth;
            originalImage.CopyPixels(originalPixels, stride, 0);

            // bitmap to be used at the end to set the image to final product
            colorPBNImage = new WriteableBitmap(origPixelWidth, origPixelHeight, 96, 96, PixelFormats.Bgra32, null);
            
            allOriginalPixels = new byte[origPixelHeight, origPixelWidth, 4];   // array of all pixels to use and manipulate

            allCells = new List<List<Pixel>>();                                 // list of cells with all of their pixels
            allPixelCellNums = new int[origPixelHeight, origPixelWidth];        // cellnumber of each pixel
            pixelMeanColor = new byte[origPixelHeight * origPixelWidth * 4, 3]; // mean color of each pixel
            byte[,] visited = new byte[origPixelHeight, origPixelWidth];        // array of whether pixel has been visited or not
            
            // copy allOriginalPixels from 1D array to 2D array and
            //  setup arrays
            int index = 0;
            for (int y = 0; y < origPixelHeight; y++)
            {
                for (int x = 0; x < origPixelWidth; x++)
                {
                    for (int i = 0; i < 4; i++)
                        allOriginalPixels[y, x, i] = originalPixels[index++];

                    visited[y, x] = (byte)UNVISITEDPIXEL;       // set pixel to unvisited
                    allPixelCellNums[y, x] = PIXELNOTINCELL;    // pixel not in cell
                }
            }

            
            // Start processing image //
            FloodFill(visited);       // group pixels into cell depending on color Euclidean Distance


            MergeSmallCells();        // merge cells with less than specific amount of pixels
            UpdateAllCells();         // update cells, get rid of empty cells
            MeanPixelColor();             // update mean color of each cell


            /*
            // Save Intermediate Images
            CreateImage();
            enc = new PngBitmapEncoder();
            enc.Frames.Add(BitmapFrame.Create(PBNImage.Source as BitmapSource));
            using (var stm = System.IO.File.Create(saveFilePath + FILENAME + saveImageNumber++ + ".png"))
                enc.Save(stm);
            */

            
            MergeLikeColorCells();  // merge cells with close color Euclidean Distance
            UpdateAllCells();       // update cells, get rid of empty cells
            MeanPixelColor();           // update mean color of each cell


            /*
            // Save Intermediate Images
            CreateImage();
            enc = new PngBitmapEncoder();
            enc.Frames.Add(BitmapFrame.Create(PBNImage.Source as BitmapSource));
            using (var stm = System.IO.File.Create(saveFilePath + FILENAME + saveImageNumber++ + ".png"))
                enc.Save(stm);
            */


            MergeToCellsWanted();         // merge cells until we get less than number of cells we want
            UpdateAllCells();   // update cells, get rid of empty cells
            MeanPixelColor();       // update mean color of each cell


            /*
            // Save Intermediate Images
            CreateImage();
            enc = new PngBitmapEncoder();
            enc.Frames.Add(BitmapFrame.Create(PBNImage.Source as BitmapSource));
            using (var stm = System.IO.File.Create(saveFilePath + FILENAME + saveImageNumber++ + ".png"))
                enc.Save(stm);
            */


            MergeToColorsWanted();                     // merge cells that have close color distance, getting specified num colors wanted
            CalculateFinalCellNumbers();    // get final cell numbers of all cells, same cellnumber for cells with same color
            FinalUpdateAllCells();          // update cells, get rid of empty cells
            MeanPixelColor();                   // update mean color of each cell with same cellnumber

            CreateImage(); // final create image
        }


        /***********************************************************************************************************
         * FloodFill to group like color pixels into cells if color EucDistance < 10                                    */
        private void FloodFill(byte[,] visited)
        {
            List<Pixel> cell;
            Stack<Pixel> pixelcellstack = new Stack<Pixel>();
            Pixel pix = new Pixel();

            byte[] start_color;

            int cellnumber = 0;

            for (int y = 0; y < origPixelHeight; y++)
            {
                for (int x = 0; x < origPixelWidth; x++)
                {
                    if (visited[y, x] == (byte)UNVISITEDPIXEL)
                    {
                        // color of first pixel
                        start_color = new byte[3];
                        for (int i = 0; i < 3; i++)
                            start_color[i] = allOriginalPixels[y, x, i];

                        
                        // create new cell starting with current pixel
                        pix.X = x;
                        pix.Y = y;
                        cell = new List<Pixel>
                        {
                            pix
                        };

                        // set pixel to visited, and it's cellnumber
                        visited[y, x] = (byte)VISITEDPIXEL;
                        allPixelCellNums[y, x] = cellnumber;

                        // add pixel to stack while it still has neighbours to look at
                        pixelcellstack.Push(pix);


                        // while there are pixels in stack to add to cell and look at neighbours
                        while (pixelcellstack.Count > 0)
                        {
                            // grab x and y values
                            int xx = pixelcellstack.Peek().X;
                            int yy = pixelcellstack.Peek().Y;

                            // LEFT
                            if (xx - 1 >= 0 && visited[yy, xx - 1] == (byte)UNVISITEDPIXEL && PixelEuclideanDistance(start_color, xx - 1, yy) < FLOOD_FILL_TOLERANCE)
                            {
                                pix.X = xx - 1;
                                pix.Y = yy;
                                pixelcellstack.Push(pix); // add neighbour to stack to look at their neighbours

                                // add neighbour pixel to cell and mark unvisited and add cellnumber
                                cell.Add(pix);
                                visited[yy, xx - 1] = (byte)VISITEDPIXEL;
                                allPixelCellNums[yy, xx - 1] = cellnumber;
                            }

                            // RIGHT
                            else if (xx + 1 < origPixelWidth && visited[yy, xx + 1] == (byte)UNVISITEDPIXEL && PixelEuclideanDistance(start_color, xx + 1, yy) < FLOOD_FILL_TOLERANCE)
                            {
                                // try new
                                pix.X = xx + 1;
                                pix.Y = yy;
                                pixelcellstack.Push(pix); // add neighbour to stack to look at their neighbours

                                // add neighbour pixel to cell and mark unvisited and add cellnumber
                                cell.Add(pix);
                                visited[yy, xx + 1] = (byte)VISITEDPIXEL;
                                allPixelCellNums[yy, xx + 1] = cellnumber;
                            }

                            // UP
                            else if (yy - 1 >= 0 && visited[yy - 1, xx] == (byte)UNVISITEDPIXEL && PixelEuclideanDistance(start_color, xx, yy - 1) < FLOOD_FILL_TOLERANCE)
                            {
                                pix.X = xx;
                                pix.Y = yy - 1;
                                pixelcellstack.Push(pix); // add neighbour to stack to look at their neighbours

                                // add neighbour pixel to cell and mark unvisited and add cellnumber
                                cell.Add(pix);
                                visited[yy - 1, xx] = (byte)VISITEDPIXEL;
                                allPixelCellNums[yy - 1, xx] = cellnumber;
                            }

                            // DOWN
                            else if (yy + 1 < origPixelHeight && visited[yy + 1, xx] == (byte)UNVISITEDPIXEL && PixelEuclideanDistance(start_color, xx, yy + 1) < FLOOD_FILL_TOLERANCE)
                            {
                                pix.X = xx;
                                pix.Y = yy + 1;
                                pixelcellstack.Push(pix); // add neighbour to stack to look at their neighbours

                                // add neighbour pixel to cell and mark unvisited and add cellnumber
                                cell.Add(pix);
                                visited[yy + 1, xx] = (byte)VISITEDPIXEL;
                                allPixelCellNums[yy + 1, xx] = cellnumber;
                            }

                            // got all neighbours, so pop pixel from stack
                            else if (pixelcellstack.Count > 0)
                                pixelcellstack.Pop();

                        }
                        // add cell to list of all cells
                        allCells.Add(cell);
                        cellnumber++; // go to next cell number
                    }
                }
            }
            // update current number of cells count
            CurrCellCount = allCells.Count;
        }
        

        /***************************************************************************
         * Merge cells with less than specified number of pixels                  */
        private void MergeSmallCells()
        {
            List<Pixel> neighbourCell;        // list of my neighbour cells
            List<int> neighbourCellNums;        // cell numbers of my neighbours
            int[] neighbourCellPixelCount;      // count of how many pixels in neighbourCell
            int neighbourCellNum;

            List<Pixel> smallCell = new List<Pixel>();  // cell with small number of pixels
            int myCellNum;                              // my cellnumber
            
            Pixel p;    // use to find pixels in neighbour cell


            // loop until no more cells with less than specific number of pixels
            for (int size = 0; size < SMALL_CELL_THRESHOLD; size++)
            {
                // as we are updating, if current amount of cells is less than cells wanted, break
                if (CurrCellCount <= NumCellsWanted)
                    return;

                // loop through all cells, only merging cells with size under threshold
                for (int i = 0; i < allCells.Count; i++)
                {
                    smallCell = allCells[i];

                    // if cell has less than specific number of pixels, merge with neighbour
                    if (smallCell.Count < size && smallCell.Count > 0)
                    {
                        myCellNum = allPixelCellNums[smallCell[0].Y, smallCell[0].X];

                        neighbourCell = new List<Pixel>();
                        neighbourCellNums = new List<int>();

                        // get neighbour allOriginalPixels of each pixel in smallCell
                        for (int j = 0; j < smallCell.Count; j++)
                        {
                            p = smallCell[j];

                            // LEFT
                            if (p.X - 1 >= 0)
                            {
                                p.X = p.X - 1;
                                neighbourCellNum = allPixelCellNums[p.Y, p.X];

                                // if cellnums are not the same and neighbourCell does not already contain pixel, add it
                                if (myCellNum != neighbourCellNum && !neighbourCell.Contains(p))
                                {
                                    neighbourCell.Add(p);

                                    // if cellnumber of neighbour not already in list of neighbourCellNums, add it
                                    if (!neighbourCellNums.Contains(neighbourCellNum))
                                        neighbourCellNums.Add(neighbourCellNum);
                                }
                            }

                            // RIGHT
                            if (p.X + 1 < origPixelWidth)
                            {
                                p.X = p.X + 1;
                                neighbourCellNum = allPixelCellNums[p.Y, p.X];

                                // if cellnums are not the same and neighbourCell does not already contain pixel, add it
                                if (myCellNum != neighbourCellNum && !neighbourCell.Contains(p))
                                {
                                    neighbourCell.Add(p);

                                    // if allPixelCellNums of neighbour not already in list of neighbourCellNums, add it
                                    if (!neighbourCellNums.Contains(neighbourCellNum))
                                        neighbourCellNums.Add(neighbourCellNum);
                                }
                            }

                            // UP
                            if (p.Y - 1 >= 0)
                            {
                                p.Y = p.Y - 1;
                                neighbourCellNum = allPixelCellNums[p.Y, p.X];

                                // if cellnums are not the same and neighbourCell does not already contain pixel, add it
                                if (myCellNum != neighbourCellNum && !neighbourCell.Contains(p))
                                {
                                    neighbourCell.Add(p);

                                    // if allPixelCellNums of neighbour not already in list of neighbourCellNums, add it
                                    if (!neighbourCellNums.Contains(neighbourCellNum))
                                        neighbourCellNums.Add(neighbourCellNum);
                                }
                            }

                            // DOWN
                            if (p.Y + 1 < origPixelHeight)
                            {
                                p.Y = p.Y + 1;
                                neighbourCellNum = allPixelCellNums[p.Y, p.X];

                                // if cellnums are not the same and neighbourCell does not already contain pixel, add it
                                if (myCellNum != neighbourCellNum && !neighbourCell.Contains(p))
                                {
                                    neighbourCell.Add(p);

                                    // if allPixelCellNums of neighbour not already in list of neighbourCellNums, add it
                                    if (!neighbourCellNums.Contains(neighbourCellNum))
                                        neighbourCellNums.Add(neighbourCellNum);
                                }
                            }
                        }

                        // count of number of pixels in each neighbour cell
                        neighbourCellPixelCount = new int[neighbourCellNums.Count];

                        foreach (Pixel P in neighbourCell)
                        {
                            for (int j = 0; j < neighbourCellNums.Count; j++)
                            {
                                // if pixel is in neighbour cell, increment count of pixels in that cell
                                if (neighbourCellNums[j] == allPixelCellNums[P.Y, P.X])
                                {
                                    neighbourCellPixelCount[j] = neighbourCellPixelCount[j] + 1;
                                    break;
                                }
                            }
                        }

                        // get neighbour cell with greatest amount of adjacent pixels
                        int largestNeighbour = 0;
                        int bestCellNum = 0;
                        for (int j = 0; j < neighbourCellPixelCount.Length; j++)
                        {
                            // if neighbour cell pixel count is greater than previous, update to this one
                            if (neighbourCellPixelCount[j] > largestNeighbour)
                            {
                                largestNeighbour = neighbourCellPixelCount[j];
                                bestCellNum = neighbourCellNums[j];
                            }

                        }

                        // merge me into best neighbour cell, clear my list
                        for (int j = 0; j < smallCell.Count; j++)
                        {
                            p = smallCell[j]; // set pixel

                            // add to neighbour cell in allCells list
                            allCells[bestCellNum].Add(p);

                            // update cellnumber of pixel
                            allPixelCellNums[p.Y, p.X] = bestCellNum;
                        }

                        allCells[myCellNum].Clear(); // clear my list
                        CurrCellCount--;             // decrement cell count

                    }

                    // break if cell count is less than number of cells wanted
                    if (CurrCellCount <= NumCellsWanted)
                        return;
                }
            }
        }

        
        /*********************************************************************************************************************************
         * Merge cells together that have a less than specified color Euclidean Distance apart                                          */
        private void MergeLikeColorCells()
        {
            // go through all cells, starting with cells with the most pixels
            List<List<Pixel>> decreasingCells = new List<List<Pixel>>();
            decreasingCells = allCells.OrderByDescending(x => x.Count).ToList();

            List<int> closeNeighbourCellNums = new List<int>();     // list of neighbour cellnumbers
            List<Pixel> toMergeNeighbourCell = new List<Pixel>();   // neighbour cell to merge to
            int neighbourCellNum;                                   // cellnumber of neighbour
            Pixel neighbour = new Pixel();                          // each neighbour pixel
            
            List<Pixel> myCell;  // cell to merge to
            int myCellNum;       // it's cellnumber

            bool MERGE = true;
            // while there are still cells to merge, distance between cell and neighbours are < CLOSE_CELL_TOLERANCE
            while (MERGE)
            {
                MERGE = false;

                for (int i = 0; i < allCells.Count; i++)
                {
                    myCell = decreasingCells[i];
                    if (myCell.Count > 0)
                    {
                        myCellNum = allPixelCellNums[myCell[0].Y, myCell[0].X];
                        myCell = allCells[myCellNum];

                        // make sure cell not already empty
                        if (allCells[myCellNum].Count > 0)
                        {
                            closeNeighbourCellNums.Clear();

                            // loop through pixels and grab neighbour cellnumbers
                            foreach (Pixel p in myCell)
                            {
                                // LEFT
                                if (p.X - 1 >= 0)
                                {
                                    neighbour.Y = p.Y;
                                    neighbour.X = p.X - 1;
                                    neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

                                    // if different cell and doesn't already contain it
                                    if (myCellNum != neighbourCellNum && !closeNeighbourCellNums.Contains(neighbourCellNum))
                                    {
                                        if (CellEuclideanDistance(myCellNum, neighbourCellNum) < CLOSE_CELL_TOLERANCE)
                                            closeNeighbourCellNums.Add(neighbourCellNum);
                                    }
                                }

                                // RIGHT
                                if (p.X + 1 < origPixelWidth)
                                {
                                    neighbour.Y = p.Y;
                                    neighbour.X = p.X + 1;
                                    neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

                                    // if different cell and doesn't already contain it
                                    if (myCellNum != neighbourCellNum && !closeNeighbourCellNums.Contains(neighbourCellNum))
                                    {
                                        if (CellEuclideanDistance(myCellNum, neighbourCellNum) < CLOSE_CELL_TOLERANCE)
                                            closeNeighbourCellNums.Add(neighbourCellNum);
                                    }
                                }

                                // UP
                                if (p.Y - 1 >= 0)
                                {
                                    neighbour.Y = p.Y - 1;
                                    neighbour.X = p.X;
                                    neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

                                    // if different cell and doesn't already contain it
                                    if (myCellNum != neighbourCellNum && !closeNeighbourCellNums.Contains(neighbourCellNum))
                                    {
                                        if (CellEuclideanDistance(myCellNum, neighbourCellNum) < CLOSE_CELL_TOLERANCE)
                                            closeNeighbourCellNums.Add(neighbourCellNum);
                                    }
                                }

                                //DOWN
                                if (p.Y + 1 < origPixelHeight)
                                {
                                    neighbour.Y = p.Y + 1;
                                    neighbour.X = p.X;
                                    neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

                                    // if different cell and doesn't already contain it
                                    if (myCellNum != neighbourCellNum && !closeNeighbourCellNums.Contains(neighbourCellNum))
                                    {
                                        if (CellEuclideanDistance(myCellNum, neighbourCellNum) < CLOSE_CELL_TOLERANCE)
                                            closeNeighbourCellNums.Add(neighbourCellNum);
                                    }
                                }
                            } // done getting neighbours

                            // if there are close neighbours, merge
                            if (closeNeighbourCellNums.Count > 0)
                            {
                                MERGE = true;

                                // merge each close neighbour cell to me and transfer their neighbour cells over
                                for (int j = 0; j < closeNeighbourCellNums.Count; j++)
                                {
                                    // loop through pixels in each neighbour and add to me
                                    toMergeNeighbourCell = allCells[closeNeighbourCellNums[j]];
                                    foreach (Pixel p in toMergeNeighbourCell)
                                    {
                                        allCells[myCellNum].Add(p);
                                        allPixelCellNums[p.Y, p.X] = myCellNum;
                                    }

                                    // update the mean cell color
                                    MeanCellColor(allCells[myCellNum]);

                                    // clear cell in allCells and update cell count
                                    toMergeNeighbourCell.Clear();
                                    CurrCellCount--;
                                }
                            }
                        }
                    }
                }
            }
        }


        /***********************************************************************************************************
         * Merge cells using small cells threshold and close cell color tolerance until less than wanted cells  */
        private void MergeToCellsWanted()
        {
            allNeighbourCells = new List<List<int>>();
            List<int> neighbourCellNums = new List<int>();
            List<int> neighbourScores = new List<int>();
            List<Pixel> myCell = new List<Pixel>();
            Pixel neighbour = new Pixel();
            int myCellNum;
            int neighbourCellNum = -1;

            bool firstGoThrough = true;

			// merge cells until we get to NumCellsWanted * 1.5
            while (CurrCellCount > NumCellsWanted * FIRST_PASS_N_RATIO)
            {
                for (int i = 0; i < allCells.Count; i++)
                {
                    myCell = allCells[i];

                    if (myCell.Count > 0)
                    {
                        myCellNum = allPixelCellNums[myCell[0].Y, myCell[0].X];

                        neighbourCellNums = new List<int>();
                        neighbourScores = new List<int>();

                        foreach (Pixel p in myCell)
                        {
                            // LEFT
                            if (p.X - 1 >= 0)
                            {
                                neighbour.Y = p.Y;
                                neighbour.X = p.X - 1;
                                neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

								// if cellnumbers of each pixel are different and neighbour is not already accounted for, add it
                                if (myCellNum != neighbourCellNum && !neighbourCellNums.Contains(neighbourCellNum))
                                {
                                    neighbourCellNums.Add(neighbourCellNum);
                                    neighbourScores.Add((int)Score(myCellNum, neighbourCellNum));
                                }
                            }

                            // RIGHT
                            if (p.X + 1 < origPixelWidth)
                            {
                                neighbour.Y = p.Y;
                                neighbour.X = p.X + 1;
                                neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

								// if cellnumbers of each pixel are different and neighbour is not already accounted for, add it
                                if (myCellNum != neighbourCellNum && !neighbourCellNums.Contains(neighbourCellNum))
                                {
                                    neighbourCellNums.Add(neighbourCellNum);
                                    neighbourScores.Add((int)Score(myCellNum, neighbourCellNum));
                                }
                            }

                            // UP
                            if (p.Y - 1 >= 0)
                            {
                                neighbour.Y = p.Y - 1;
                                neighbour.X = p.X;
                                neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

								// if cellnumbers of each pixel are different and neighbour is not already accounted for, add it
                                if (myCellNum != neighbourCellNum && !neighbourCellNums.Contains(neighbourCellNum))
                                {
                                    neighbourCellNums.Add(neighbourCellNum);
                                    neighbourScores.Add((int)Score(myCellNum, neighbourCellNum));
                                }
                            }

                            //DOWN
                            if (p.Y + 1 < origPixelHeight)
                            {
                                neighbour.Y = p.Y + 1;
                                neighbour.X = p.X;
                                neighbourCellNum = allPixelCellNums[neighbour.Y, neighbour.X];

								// if cellnumbers of each pixel are different and neighbour is not already accounted for, add it
                                if (myCellNum != neighbourCellNum && !neighbourCellNums.Contains(neighbourCellNum))
                                {
                                    neighbourCellNums.Add(neighbourCellNum);
                                    neighbourScores.Add((int)Score(myCellNum, neighbourCellNum));
                                }
                            }
                        }

						
						// go through all neighbours and find the one with lowest score
                        int min = -1, minNeighbourCellNum = 0, minMyCellNum = 0;
                        for (int j = 0; j < neighbourScores.Count; j++)
                        {
                            int value = neighbourScores[j];
                            if (min == -1 || value < min)
                            {
                                min = value;
                                minNeighbourCellNum = neighbourCellNums[j];
                                minMyCellNum = myCellNum;
                            }
                        }


                        // now merge cells together and clear neighbour cell
                        Pixel pix = new Pixel();
                        for (int j = 0; j < allCells[minNeighbourCellNum].Count; j++)
                        {
                            pix = allCells[minNeighbourCellNum][j];

                            allCells[minMyCellNum].Add(pix);
                            allPixelCellNums[pix.Y, pix.X] = minMyCellNum;
                        }

                        MeanCellColor(allCells[minMyCellNum]); // update mean color of new cell

                        allCells[minNeighbourCellNum].Clear(); // clear neighbour cell

                        CurrCellCount--; // update cell count
                        
                        neighbourCellNums.Remove(minNeighbourCellNum); // remove the neighbour cellnumber
                        if (firstGoThrough) // add all neighbours
                        {
                            allNeighbourCells.Add(neighbourCellNums);
                        }
                    }
                }
                firstGoThrough = false;
            }

        }
        private double Score (int cellnum1, int cellnum2)
        {
            return CellEuclideanDistance(cellnum1, cellnum2) * Math.Pow(allCells[cellnum1].Count, .5);
        }
        

        /**********************************************************************************************************
         * Merge cells until number of colors wanted is achieved. 30 times to get the best random mix of colors  */
        private void MergeToColorsWanted()
        {
            List<List<int>> best_cellGroup = new List<List<int>>();
            int best_score = -1;

            List<byte[]> colorGroup = new List<byte[]>();

            List<int> cellGroup;
            byte[] meancolorbyte = new byte[3];
            Random rnd = new Random();

            bool converged;

            // loop through, using k means this many times to get best color group
            for (int i = 0; i < K_MEANS_TRIALS; i++)
            {
                // create groups //
                colorGroup = new List<byte[]>();
                cellGroup = new List<int>();
                for (int j = 0; j < NumColorsWanted; j++)
                {
                    // grab random cellnumbers
                    int cellnumber = rnd.Next(allCells.Count);
                    while (cellGroup.Contains(cellnumber))
                        cellnumber = rnd.Next(allCells.Count);

                    cellGroup.Add(cellnumber);

                    byte[] color = new byte[3];
                    for (int l = 0; l < 3; l++)
                        color[l] = pixelMeanColor[cellnumber, l];
                    colorGroup.Add(color);
                }

				// keep getting groups of cells until best group is determined //
                converged = false;
                while (!converged)
                {
                    List<List<int>> clusters = new List<List<int>>(new List<int>[cellGroup.Count]);
                    for (int j = 0; j < clusters.Count; j++)
                        clusters[j] = new List<int>();

                    FormGroups(cellGroup, clusters);
                    
                    List<byte[]> newColorGroup = new List<byte[]>();
                    foreach (List<int> clusterCellNums in clusters)
                    {
                        // returns color for the cluster at centroid
                        newColorGroup.Add(CalculateColor(clusterCellNums));
                    }
                    
                    converged = true;
                    for (int j = 0; j < colorGroup.Count; j++)
                    {
                        for (int l = 0; l < 3; l++)
                        {
                            if (colorGroup[j][l] != newColorGroup[j][l])
                            {
                                converged = false;
                                break;
                            }
                        }
                    }

                    colorGroup = newColorGroup;

                    // continually getting the best group of cells
                    int score = CalculateScore(clusters, cellGroup);
                    if (best_score == -1 || score < best_score)
                    {
                        best_score = score;
                        best_cellGroup = clusters;
                    }
                }
            }

            // update mean color of each cell in group
            for (int i = 0; i < best_cellGroup.Count; i++)
            {
                foreach (int cellnumber in best_cellGroup[i])
                {
                    for (int j = 0; j < 3; j++)
                        pixelMeanColor[cellnumber, j] = colorGroup[i][j];
                }
            }
        }
		
		// Form groups of cells depending on distance between new cell and the group of randomly chosen cells
        public void FormGroups(List<int> centroids, List<List<int>> clusters)
        {
            // each centroid is a cellnumber
            foreach (List<Pixel> cell in allCells)
            {
                // add cell to closest centroid
                int oldscore;
                int newscore;
                int closest_centroid = 0;

                oldscore = CellEuclideanDistance(allPixelCellNums[cell[0].Y, cell[0].X], centroids[0]);
                for (int i = 1; i < centroids.Count; i++)
                {
                    // celleuclideandistance(myCellNum, neighbourCellNum, pixelMeanColor);
                    newscore = CellEuclideanDistance(allPixelCellNums[cell[0].Y, cell[0].X], centroids[i]);

                    if (newscore < oldscore)
                    {
                        oldscore = newscore;
                        closest_centroid = i;
                    }
                }

                // oldscore and closest_centroid is centroidnumber and distance of closest
                clusters[closest_centroid].Add(allPixelCellNums[cell[0].Y, cell[0].X]);
            }
        }
		
		// Calculate the mean color for each cluster of cells
        public byte[] CalculateColor(List<int> clusterCellNums)
        {
            double B_sum = 0, G_sum = 0, R_sum = 0;
            double weighting = 0;

            // cellnumber of each cell in cluster
            for (int i = 0; i < clusterCellNums.Count; i++)
            {
                // cellnumber is clusterCellNums[i]
                int cellnumber = clusterCellNums[i];

                byte[] color = new byte[3];
                for (int l = 0; l < 3; l++)
                    color[l] = pixelMeanColor[cellnumber, l];

                // weight based on cell size
                double cell_weight = Math.Pow(allCells[cellnumber].Count, .5);

                B_sum += color[0] * cell_weight;
                G_sum += color[1] * cell_weight;
                R_sum += color[2] * cell_weight;

                weighting += cell_weight;

            }
            byte[] colorcluster = new byte[3];
            colorcluster[0] = (byte)(B_sum / weighting);
            colorcluster[1] = (byte)(G_sum / weighting);
            colorcluster[2] = (byte)(R_sum / weighting);

            return colorcluster;
        }
		
		// Calculate score for each cluster of cells/colors
        public int CalculateScore(List<List<int>> clusters, List<int> centroids)
        {
            List<double> scatter = new List<double>(new double[centroids.Count]);

            double scatterscore;

            List<int> cluster;

            int centroid1_cellnumber;
            int centroid2_cellnumber;
            int distance;
            int cluster_cellnumber;
            double index;

            double[] distance_scores;

            for (int i = 0; i < centroids.Count; i++)
            {
                centroid1_cellnumber = centroids[i];
                cluster = clusters[i];
                scatterscore = 0;

                for (int j = 0; j < cluster.Count; j++)
                {
                    cluster_cellnumber = cluster[j];

                    distance = CellEuclideanDistance(centroid1_cellnumber, cluster_cellnumber);

                    scatterscore += distance * Math.Pow(allCells[cluster_cellnumber].Count, .5);
                }

                scatterscore /= cluster.Count;
                scatter[i] = Math.Pow(scatterscore, 2);
            }


			// for each centroid, check for other matching centroid with matching color
            index = 0;
            for (int i = 0; i < centroids.Count; i++)
            {
                centroid1_cellnumber = centroids[i];

                distance_scores = new double[centroids.Count];

                for (int j = 0; j < centroids.Count; j++)
                {
                    centroid2_cellnumber = centroids[j];

					// if cells have same mean color, update score of cell
                    if (ColorMatch(centroid1_cellnumber, centroid2_cellnumber) == false)
                    {
                        double score = (scatter[i] + scatter[j]) / CellEuclideanDistance(centroid1_cellnumber, centroid2_cellnumber);

                        distance_scores[j] = score;
                    }
                }

                index += distance_scores.Max(); // add the max of the scores to index
            }

            return (int)index; // return index
        }
		
		// Returns true if mean colors of cells sent are equal, else returns false
        public bool ColorMatch(int cellnumber1, int cellnumber2)
        {
            bool colorMatch = true;
            for (int i = 0; i < 3; i++)
            {
                if (pixelMeanColor[cellnumber1, 0] != pixelMeanColor[cellnumber2, 0] && pixelMeanColor[cellnumber1, 1] != pixelMeanColor[cellnumber2, 1] && pixelMeanColor[cellnumber1, 2] != pixelMeanColor[cellnumber2, 2])
                    colorMatch = false;
            }

            return colorMatch;
        }


        /*************************************************************************************
         * EucDistance to return distance between colors of two pixels                      */
        private int PixelEuclideanDistance(byte[] start_color, int x, int y)
        {
            double blue = Math.Pow(allOriginalPixels[y, x, 0] - start_color[0], 2);
            double green = Math.Pow(allOriginalPixels[y, x, 1] - start_color[1], 2);
            double red = Math.Pow(allOriginalPixels[y, x, 2] - start_color[2], 2);

            return (int)Math.Sqrt(blue + green + red);
        }


        /*******************************************************************************************
         * EucDistance to return distance between meancolors of two cells                         */
        private int CellEuclideanDistance(int myCellNum, int neighbourCellNum)
        {
            double blue = Math.Pow(pixelMeanColor[neighbourCellNum, 0] - pixelMeanColor[myCellNum, 0], 2);
            double green = Math.Pow(pixelMeanColor[neighbourCellNum, 1] - pixelMeanColor[myCellNum, 1], 2);
            double red = Math.Pow(pixelMeanColor[neighbourCellNum, 2] - pixelMeanColor[myCellNum, 2], 2);

            return (int)Math.Sqrt(blue + green + red);
        }


        /***********************************************************************************************************
         * Calculate mean colors of all cells using the pixels in each cell                                                 */
        private void MeanPixelColor()
        {
            int B_sum, G_sum, R_sum;
            foreach (List<Pixel> currentcell in allCells)
            {
                if (currentcell.Count > 0)
                {
                    B_sum = 0;
                    G_sum = 0;
                    R_sum = 0;

                    foreach (Pixel p in currentcell)
                    {
                        B_sum += allOriginalPixels[p.Y, p.X, 0];
                        G_sum += allOriginalPixels[p.Y, p.X, 1];
                        R_sum += allOriginalPixels[p.Y, p.X, 2];
                    }

                    // set mean color for each cellnumber
                    int currcellnum = allPixelCellNums[currentcell[0].Y, currentcell[0].X];
                    pixelMeanColor[currcellnum, 0] = (byte)(B_sum / currentcell.Count);
                    pixelMeanColor[currcellnum, 1] = (byte)(G_sum / currentcell.Count);
                    pixelMeanColor[currcellnum, 2] = (byte)(R_sum / currentcell.Count);
                }
            }
        }


        /****************************************************************************************************
         * Calculate mean color of cell using all pixels in cell                                              */
        private void MeanCellColor(List<Pixel> cell)
        {
			// recalculate mean color of each cell using color of each pixel in cell
			
            int B_sum = 0;
            int G_sum = 0;
            int R_sum = 0;

            foreach (Pixel p in cell)
            {
                B_sum += allOriginalPixels[p.Y, p.X, 0];
                G_sum += allOriginalPixels[p.Y, p.X, 1];
                R_sum += allOriginalPixels[p.Y, p.X, 2];
            }

            // update mean color of each cellnumber
            int currcellnum = allPixelCellNums[cell[0].Y, cell[0].X];
            pixelMeanColor[currcellnum, 0] = (byte)(B_sum / cell.Count);
            pixelMeanColor[currcellnum, 1] = (byte)(G_sum / cell.Count);
            pixelMeanColor[currcellnum, 2] = (byte)(R_sum / cell.Count);
        }


        /*******************************************************************************
         * Update allCells list so that there are no empty cells                      */
        private void UpdateAllCells()
        {
			// remove all empty cells
            List<Pixel> cell;
            for (int i = 0; i < allCells.Count; i++)
            {
                while (i < allCells.Count && allCells[i].Count == 0)
                    allCells.RemoveAt(i);
            }

			// update cellnumbers of all the pixels
            for (int i = 0; i < allCells.Count; i++)
            {
                cell = allCells[i];
                for (int j = 0; j < cell.Count; j++)
                    allPixelCellNums[cell[j].Y, cell[j].X] = i;
            }
            
        }

        
        /*****************************************************************************
         * Updates all cells to have correct cellnumbers and meancolors             */
        private void CalculateFinalCellNumbers()
        {
			// keep copy of all separate cells
            finalAllCells = allCells.ConvertAll(cell => new List<Pixel>());
            for (int i = 0; i < allCells.Count; i++)
            {
                for (int j = 0; j < allCells[i].Count; j++)
                    finalAllCells[i].Add(allCells[i][j]);
            }

			// to keep track of cells mapped to their new final cellnumbers
            cellMapCellNumber = new int[allCells.Count];

			// go through all cells, updating their cellnumbers, combining cells to have same colors
            for (int i = 0; i < allCells.Count; i++)
            {
                if (allCells[i].Count > 0)
                {
                    byte[] color = new byte[3];
                    for (int l = 0; l < 3; l++)
                        color[l] = pixelMeanColor[i, l];

                    for (int j = i + 1; j < allCells.Count; j++)
                    {
                        if (allCells[j].Count > 0)
                        {

                            byte[] newcolor = new byte[3];
                            for (int l = 0; l < 3; l++)
                                newcolor[l] = pixelMeanColor[j, l];

							// if cells are close in color, combine them and map to new cellnumber
                            if (CellEuclideanDistance(i, j) < 10)
                            {
                                foreach (Pixel p in allCells[j])
                                {
                                    allPixelCellNums[p.Y, p.X] = i;
                                    
                                    allCells[i].Add(p);
                                }
                                
                                allCells[j].Clear();

                                // update mapped cellnumber of j to i
                                cellMapCellNumber[j] = i;
                            }
                            else
                            {
                                // set mapped cellnumber of j to j
                                cellMapCellNumber[j] = j;
                            }
                        }
                    }
                }
            }
        }


        /*******************************************************************************
         * Last update on allCells and cellnumbers after merging to colors wanted     */
        private void FinalUpdateAllCells()
        {
            // update all cells, removing empty cells and updating cellnumbers
            bool empty = true;
            while (empty == true)
            {
                empty = false;
                for (int i = 0; i < allCells.Count; i++)
                {
                    if (allCells[i].Count == 0)
                    {
                        for (int j = 0; j < cellMapCellNumber.Count(); j++)
                        {
                            if (cellMapCellNumber[j] > i)
                                cellMapCellNumber[j]--;
                        }
                        
                        allCells.RemoveAt(i);
                        i--;

                        empty = true;
                    }
                }
            }
            
			// now update cellnumbers of each pixel
			List<Pixel> cell;
            for (int i = 0; i < allCells.Count; i++)
            {
                cell = allCells[i];
                for (int j = 0; j < cell.Count; j++)
                    allPixelCellNums[cell[j].Y, cell[j].X] = i;
            }
            
        }


        /*****************************************************************************************************
         * Creates colorPBNImage image so it can be outputted onto canvas and saved intermittently               */
        private void CreateImage()
        {
			// set each pixel to it's cell mean color
            byte blue, green, red;
            for (int y = 0; y < origPixelHeight; y++)
            {
                for (int x = 0; x < origPixelWidth; x++)
                {
                    blue = pixelMeanColor[allPixelCellNums[y, x], 0];
                    green = pixelMeanColor[allPixelCellNums[y, x], 1];
                    red = pixelMeanColor[allPixelCellNums[y, x], 2];
                    allOriginalPixels[y, x, 0] = blue;
                    allOriginalPixels[y, x, 1] = green;
                    allOriginalPixels[y, x, 2] = red;

                    allOriginalPixels[y, x, 3] = 255;
                }
            }

			// turn 2D array back into 1D array
            byte[] pixels1D = new byte[origPixelHeight * origPixelWidth * 4];
            int ind = 0;
            for (int y = 0; y < origPixelHeight; y++)
            {
                for (int x = 0; x < origPixelWidth; x++)
                {
                    for (int i = 0; i < 4; i++)
                        pixels1D[ind++] = allOriginalPixels[y, x, i];
                }
            }

			// create new bitmap and display in image on canvas
            Int32Rect rect = new Int32Rect(0, 0, origPixelWidth, origPixelHeight);
            int str = 4 * origPixelWidth;
            colorPBNImage.WritePixels(rect, pixels1D, str, 0);

            PBNImage.Source = colorPBNImage;
        }
    }
}